@define-color polari_dark_bg_color shade(@theme_bg_color, 0.95);

.polari-room-list .sidebar {
    border-style: solid;
    border-color: @borders;
    border-image: none;
}

.polari-room-list .sidebar:dir(ltr) {
    border-width: 0 1px 0 0;
}

.polari-room-list .sidebar:dir(rtl) {
    border-width: 0 0 0 1px;
}

.polari-user-list row.expanded {
    background-color: @content_view_bg;
}

.polari-user-list row frame {
    box-shadow: inset 0 2px 4px alpha(@polari_dark_bg_color, 0.8);
    transition: all 250ms ease-out;
    border-width: 0;
}

.polari-paste-confirmation {
    border: 0 solid @borders;
    border-top-width: 1px;
}

.polari-nick-button {
    padding-left: 8px;
    padding-right: 8px;
}

.polari-nick-entry {
    color: alpha(@theme_fg_color, 0.6);
}

.polari-room-list .room-list-header {
    font-size: smaller;
    padding: 8px 7px 0;
}

.polari-room-list .room-list-header label:dir(ltr),
.polari-room-list .room-list-header stack:dir(rtl) {
    padding-left: 8px;
}

.polari-room-list .room-list-header stack:dir(ltr),
.polari-room-list .room-list-header label:dir(rtl) {
    padding-right: 8px;
}

.polari-room-list .room-list-header separator {
    background-color: alpha(@borders, 0.55);
}

.polari-room-list row {
    padding-left: 8px;
}

.polari-room-list row.inactive *,
.polari-room-list row.inactive:backdrop * {
    color: mix(@theme_fg_color, @theme_base_color, 0.3);
}

.polari-room-list row.inactive:selected *,
.polari-room-list row.inactive:selected:backdrop * {
    color: mix(@theme_selected_fg_color, @theme_base_color, 0.3);
}

.polari-room-list row:backdrop * {
    color: @theme_fg_color;
}

.polari-room-list row:backdrop:selected * {
    color: @theme_selected_fg_color;
}

.polari-room-list row .pending-messages-count {
    color: @theme_selected_fg_color;
    background-color: mix(@theme_selected_bg_color, @theme_selected_fg_color, 0.3);
    font-size: smaller;
    border-radius: 4px;
    padding: 0px 4px;
}

.polari-room-list row .pending-messages-count:backdrop {
    color: @theme_unfocused_selected_fg_color;
    background-color: mix(@theme_unfocused_selected_bg_color, @theme_unfocused_selected_fg_color, 0.1);
}

.irc-feedback {
    color: @theme_fg_color;
    background-color: @theme_bg_color;
    border-style: solid;
    border-color: @borders;
    border-width: 1px 1px 0 1px;
    border-radius: 6px 6px 0 0;
    padding: 6px;
}

.titlebar headerbar:dir(ltr):first-child,
.titlebar headerbar:dir(rtl):last-child {
    /* the horizontal padding confuses left_side_size_group, so we
       add it as margin-left to the first button instead (in mainWindow.ui);
       also set the vertical padding to 0 so the separator extends to the
       bottom */
    padding: 0 1px;
}

.polari-titlebar-separator {
    border-image: linear-gradient(to bottom, @theme_bg_color, shade(@borders, 1.2));
    border-image-width: 0 1px 0 0;
    border-image-slice: 1;
}

.polari-titlebar-separator:backdrop {
    border-image: none;
    border-color: transparent;
}

.polari-user-list-button {
    padding: 0px 4px;
}

.polari-background {
    opacity: 0.5;
}

.polari-background-title {
    font-size: 250%;
}

.polari-background-description {
    font-size: larger;
}

.polari-error-information {
    background-color: @polari_dark_bg_color;
    border-style: solid;
    border-color: @borders;
    border-width: 1px 0 0 0;
}
.polari-error-information GtkLabel {
    color: @theme_fg_color;
    font-size: smaller;
}

.polari-listbox-filterbar {
    border: solid @borders;
    border-width: 0 0 1px;
}
